import numpy as np
from girard import linear_algebra as la

def sample_hypersphere_point(dim):
    gaussianSamples = np.random.normal(0, 1, dim)
    return la.normalize(gaussianSamples)


def sample_cone_uniformly_from(dimension):
    np.matrix([sample_hypersphere_point(dimension) for i in range(dimension)]).T



def _get_version() -> str:
    """
    Get package version.

    Returns
    -------
    version : str
    """
    from os.path import dirname, join
    with open(join(dirname(__file__), "resources", "VERSION")) as f:
        return f.read().strip()


__version__ = _get_version()



from decimal import Decimal

memoization = {
    "factorial": {"cache": {0: Decimal(1), 1: Decimal(1)},
                  "step": 1, "factor": lambda x: Decimal(x),
                  "max": 1},
    "odd_factorial": {"cache": {2: Decimal(1), 4: Decimal(3)},
                      "step": 2, "factor": lambda x: Decimal(x - 1),
                      "max": 4},
    "power_of_two": {"cache": {0: Decimal(1), 1: Decimal(2)},
                     "step": 1, "factor": lambda x: Decimal(2),
                     "max": 1}
}


def _generic_memoized_fn(n, fn_type):
    cache, step, factor, max_memoized = memoization[fn_type].values()

    if n in cache:
        return cache[n]

    for value in range(max_memoized, n + 1, step):
        fn = factor(value) * cache[value - step]
        cache[value] = fn

    return cache[n]


def factorial(n):
    return _generic_memoized_fn(n, fn_type="factorial")


def power_of_two(n):
    return _generic_memoized_fn(n, fn_type="power_of_two")


def odd_factorial(n):
    assert n % 2 == 0, "Input must be an even number"
    return _generic_memoized_fn(n, fn_type="odd_factorial")



import numpy as np

from math import pi, sqrt, exp
from functools import reduce
from numpy import matrix, arccos
from numpy.linalg import inv, det
from decimal import *

from girard import linear_algebra as la, memoize as mem, utils, series_convergence as sc

#term of the hypergeometric series in dimension \binom{n}{k}.
# a and alpha are vectors in dimension \binom{n}{k}
def series_term(m_tuple, alpha, n):
    alpha_to_the_m = la.vector_vetor_exp(alpha, m_tuple)
    sum_entries = sum(m_tuple)
    sign = (-1) ** sum_entries
    power = mem.power_of_two(sum_entries)
    fact = reduce(lambda a,b: a*b, map(mem.factorial, m_tuple))
    gamma_prod = utils.gamma_product(m_tuple, n)
    return (alpha_to_the_m * sign * power * gamma_prod) / fact

# function that puts together all previous components to implement Ribando"s formula for the solid angle
def solid_angle(spanning_matrix, eps, max_weight=50):
    cone_vectors = list(map(la.normalize, spanning_matrix.T.tolist()))

    dimension = len(cone_vectors)

    alpha = la.pairwise_dot_products(cone_vectors)
    normalized_spanning_matrix = np.matrix(cone_vectors).T

    series_will_converge = sc.check_convergence(normalized_spanning_matrix)

    if not series_will_converge:
        raise ValueError("The cone inputed is not within the region of convergence of the hypergeometric series. Use the monte_carlo module instead.")

    det_V = abs(np.linalg.det(normalized_spanning_matrix))

    pi_term = (4*np.pi)**(dimension/2.0)
    constant_term = Decimal(det_V / pi_term)

    series_dimension = int(dimension*(dimension-1)/2)
    inputs = utils.tuple_generator(series_dimension)
    number_of_tuples = utils.number_of_d_tuples_with_max_weight_k(series_dimension, max_weight)

    approx = 0
    its = 0
    while its < number_of_tuples:
        term = series_term(next(inputs), alpha, dimension)
        approx += term
        its += 1
        if abs(term) < eps:
            break

    return constant_term * approx



import numpy as np

from math import sqrt
from numpy import linalg as npla
from scipy.optimize import linprog


def diagonalize_symmetric_matrix(M):
    eigenvalues, eigenvectors = npla.eig(M)

    diagonal_matrix = np.matrix(np.diagflat(eigenvalues))
    orthogonal_matrix = eigenvectors.T # the columns of orthogonal_matrix are the eigenvectors

    return orthogonal_matrix, diagonal_matrix


def square_root_positive_semidefinite_matrix(M):
    eigenvalues, orthogonal_eigenvectors = npla.eig(M)

    is_positive_semidefinite = all([x >= 0 for x in eigenvalues])

    if not is_positive_semidefinite:
        raise ValueError('The matrix must be positive semidefinite (have non-negative eigenvalues)')

    diagonal_square_root = np.matrix(np.diagflat([sqrt(x) for x in eigenvalues]))

    return orthogonal_eigenvectors * diagonal_square_root * orthogonal_eigenvectors.T


def flip_coin():
    return np.random.rand() > 0.5


def ones_vec(d):
    return np.matrix([1 for i in range(d)]).T


def fill_vertex_neighborhood(vertex_label, adjacency_matrix):
    neighborhood = adjacency_matrix[vertex_label].A1

    while True:
        for idx in range(adjacency_matrix.shape[0]):
            if idx != vertex_label and neighborhood[idx] == 0 and flip_coin():
                adjacency_matrix[vertex_label, idx] = adjacency_matrix[idx, vertex_label] = 1

        if sum(neighborhood) >= 2:
            break


def generate_graph(d):
    adjacency_matrix = np.matrix(np.zeros((d,d)))

    for label in range(d):
        fill_vertex_neighborhood(label, adjacency_matrix)

    return adjacency_matrix


def build_linear_constraints(graph):
    nvertices = graph.shape[0]
    linearlized_index = 0
    variable_mapping = {}
    rows = []

    for i in range(nvertices):
        for j in range(i + 1, nvertices):
            if i != j and graph[i, j] == 1:
                variable_mapping[linearlized_index] = (i, j)
                linearlized_index += 1

    for i in range(nvertices):
        rows.append([1 if i in variable_mapping[j] else 0 for j in range(linearlized_index)])

    return np.matrix(rows), variable_mapping


def generate_random_weighted_graph(d):
    g = generate_graph(d)
    linear_constraints, variable_mapping = build_linear_constraints(g)
    nedges = len(variable_mapping)

    random_objective_function = np.matrix([int(x * 10) + 1 for x in np.random.rand(nedges)]).T
    all_ones_obj = ones_vec(nedges)

    weights = linprog(c=all_ones_obj, A_eq=linear_constraints, b_eq=ones_vec(d)).x

    for linearlized_index, vertex_pair in variable_mapping.items():
        v = vertex_pair[0]
        u = vertex_pair[1]
        g[v, u] = g[u, v] = weights[linearlized_index]

    return g


def Identity(d):
    return np.matrix(np.diagflat([1 for i in range(d)]))


def trim_laplacian(g):
    g2 = g.copy()

    g2 = np.delete(g2, 0, axis=0)
    g2 = np.delete(g2, 0, axis=1)

    return g2


def generate_random_convergent_cone(d):
    G = generate_random_weighted_graph(d + 1)
    L = Identity(d + 1) - G
    T = trim_laplacian(L)
    VTV = 2*Identity(d) - T
    return square_root_positive_semidefinite_matrix(VTV)


def check(d):
    V = generate_random_convergent_cone(d)
    VTV = V.T * V
    M = 2 * Identity(d) - VTV
    return min(np.linalg.eigvals(M))
